from abc import ABC, abstractmethod

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


class BaseProcessor(ABC):
    display_name = None
    accepted_currencies = None
    logo_url = None
    slug = None  # for friendly urls
    method = "GET"
    template_name = None

    def __init__(self, payment):
        self.payment = payment
        self.path = payment.backend
        self.context = {}  # can be used by Payment's customized methods.
        if self.slug is None:
            self.slug = payment.backend  # no more Mr. Friendly :P
        self.config = getattr(settings, "GETPAID_BACKEND_SETTINGS", {}).get(
            self.path, {}
        )

    def get_form(self, post_data):
        """
        Only used if the payment processor requires POST requests.
        Generates a form only containing hidden input fields.
        """
        from . import forms

        return forms.PaymentHiddenInputsPostForm(items=post_data)

    def handle_callback(self, request, *args, **kwargs):
        """
        This method handles the callback from payment broker for the purpose
        of updating the payment status in our system.
        :param args:
        :param kwargs:
        :return: HttpResponse instance
        """
        raise NotImplementedError

    @classmethod
    def get_display_name(cls):
        return cls.display_name

    @classmethod
    def get_accepted_currencies(cls):
        return cls.accepted_currencies

    @classmethod
    def get_logo_url(cls):
        return cls.logo_url

    def fetch_status(self):
        """
        Logic for checking payment status with broker.

        Should return dict with either "amount" or "status" keys.
        If "status" key is used, it should be one of getpaid.models.PAYMENT_STATUS_CHOICES
        If both keys are present, "status" takes precedence.
        """
        raise NotImplementedError

    @abstractmethod
    def get_redirect_params(self):
        return {}

    def get_redirect_method(self):
        return self.method

    @abstractmethod
    def get_redirect_url(self):
        return

    def get_template_names(self, view=None):
        template_name = self.get_setting("POST_TEMPLATE")
        if template_name is None:
            template_name = getattr(settings, "GETPAID_POST_TEMPLATE", None)
        if template_name is None:
            template_name = self.template_name
        if template_name is None and hasattr(view, "get_template_names"):
            return view.get_template_names()
        if template_name is None:
            raise ImproperlyConfigured("Couldn't determine template name!")
        return [template_name]

    def get_setting(self, name, default=None):
        return self.config.get(name, default)



import json

from django.http import HttpResponse
from django.urls import reverse

from getpaid.processor import BaseProcessor


class PaymentProcessor(BaseProcessor):
    slug = "dummy"
    display_name = "Dummy"
    accepted_currencies = ["PLN", "EUR"]
    method = "POST"
    template_name = "getpaid_dummy_backend/payment_post_form.html"

    def handle_callback(self, request, *args, **kwargs):
        payload = json.loads(request.data)
        if payload["status"] == "OK":
            self.payment.on_success()
        else:
            self.payment.on_failure()

        return HttpResponse("OK")

    def get_redirect_params(self):
        extra_args = {}
        if self.get_setting("confirmation_method", "push").lower() == "push":
            extra_args["callback"] = (
                reverse("getpaid:callback-detail", kwargs=dict(pk=self.payment.pk)),
            )

        return dict(
            payment=self.payment.pk,
            value=self.payment.amount,
            currency=self.payment.currency,
            description=self.payment.description,
            success_url=reverse(
                "getpaid:payment-success", kwargs=dict(pk=self.payment.pk)
            ),
            failure_url=reverse(
                "getpaid:payment-failure", kwargs=dict(pk=self.payment.pk)
            ),
            **extra_args,
        )

    def get_redirect_url(self):
        return reverse("getpaid:dummy:gateway")



from django.conf.urls import include, url

from . import views
from .registry import registry

app_name = "getpaid"

urlpatterns = [
    url(
        r"^new/(?P<currency>[A-Z]{3})/$",
        views.CreatePaymentView.as_view(),
        name="create-payment",
    ),
    url(
        r"^success/(?P<pk>[0-9a-f-]+)/$",
        views.SuccessView.as_view(),
        name="payment-success",
    ),
    url(
        r"^failure/(?P<pk>[0-9a-f-]+)/$",
        views.FailureView.as_view(),
        name="payment-failure",
    ),
    url(
        r"^callback/(?P<pk>[0-9a-f-]+)/$",
        views.CallbackView.as_view(),
        name="callback-detail",
    ),
    url(r"^backends/", include(registry.urls)),
]



from django import forms
from django.utils.translation import ugettext as _


class DummyQuestionForm(forms.Form):
    """
    This dummy form asks for payment authorization.
    """

    authorize_payment = forms.ChoiceField(
        label=_("authorization"), choices=((1, _("yes")), (0, _("no")))
    )

    callback = forms.CharField(widget=forms.HiddenInput)
    success_url = forms.CharField(widget=forms.HiddenInput)
    failure_url = forms.CharField(widget=forms.HiddenInput)



import importlib

from django.conf.urls import include, url

from getpaid.processor import BaseProcessor


def importable(name):
    try:
        importlib.import_module(name)
        return True
    except ImportError:
        return False


class PluginRegistry(object):
    def __init__(self):
        self._backends = {}

    def __contains__(self, item):
        return item in self._backends

    def __getitem__(self, item):
        return self._backends[item]

    def __iter__(self):
        return iter(self._backends)

    def register(self, module_or_proc):
        if hasattr(module_or_proc, "__base__") and issubclass(
            module_or_proc, BaseProcessor
        ):
            self._backends[module_or_proc.slug] = module_or_proc
        else:
            self._backends[
                module_or_proc.__name__
            ] = module_or_proc.processor.PaymentProcessor

    def get_choices(self, currency):
        currency = currency.upper()
        return [
            (name, p.display_name)
            for name, p in self._backends.items()
            if currency in p.get_accepted_currencies()
        ]

    @property
    def urls(self):
        return [
            url(
                r"^{}/".format(p.slug),
                include(("{}.urls".format(name), p.slug), namespace=p.slug),
            )
            for name, p in self._backends.items()
            if importable("{}.urls".format(name))
        ]

    def get_all_supported_currency_choices(self):
        currencies = set()
        for backend in self._backends:
            currencies.update(backend.accepted_currencies or [])
        return [(c.upper(), c.upper()) for c in currencies]


registry = PluginRegistry()



import uuid
from importlib import import_module

import pendulum
import swapper
from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _

from . import FraudStatus, PaymentStatus, signals
from .registry import registry


class AbstractOrder(models.Model):
    class Meta:
        abstract = True

    def get_redirect_url(self, *args, success=None, **kwargs):
        """
        Method used to determine the final url the client should see after
        returning from gateway. Client will be redirected to this url after
        backend handled the original callback (i.e. updated payment status)
        and only if SUCCESS_URL or FAILURE_URL settings are NOT set.
        By default it returns the result of `get_absolute_url`
        """
        return self.get_absolute_url()

    def get_absolute_url(self):
        """
        Standard method recommended in Django docs. It should return
        the URL to see details of particular Order.
        """
        raise NotImplementedError

    def is_ready_for_payment(self):
        """Most of the validation is made in PaymentMethodForm using but if you
        need any extra validation. For example you most probably want to disable
        making another payment for order that is already paid."""
        return True

    def get_items(self):
        """
        There are backends that require some sort of item list to be attached
        to the payment. But it's up to you if the list is real or contains only
        one item called "Payment for stuff in {myshop}" ;)
        :return: List of {"name": str, "quantity": Decimal, "unit_price": Decimal} dicts.
        """
        return [
            {
                "name": self.get_description(),
                "quantity": 1,
                "unit_price": self.get_total_amount(),
            }
        ]

    def get_total_amount(self):
        """
        This method must return the total value of the Order.
        :return: Decimal object
        """
        raise NotImplementedError

    def get_user_info(self):
        """
        This method should return dict with necessary user info.
        For most backends email should be sufficient.
        Expected field names: `email`, `first_name`, `last_name`, `phone`
        """
        raise NotImplementedError

    def get_description(self):
        """
        :return: Description of the Order. Should return the value of appropriate field.
        """
        raise NotImplementedError


class AbstractPayment(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    order = models.ForeignKey(
        swapper.get_model_name("getpaid", "Order"),
        verbose_name=_("order"),
        on_delete=models.CASCADE,
        related_name="payments",
    )
    amount = models.DecimalField(_("amount"), decimal_places=4, max_digits=20)
    currency = models.CharField(_("currency"), max_length=3)
    status = models.CharField(
        _("status"),
        max_length=20,
        choices=PaymentStatus.CHOICES,
        default=PaymentStatus.NEW,
        db_index=True,
    )
    backend = models.CharField(_("backend"), max_length=50)
    created_on = models.DateTimeField(_("created on"), auto_now_add=True, db_index=True)
    paid_on = models.DateTimeField(
        _("paid on"), blank=True, null=True, default=None, db_index=True
    )
    amount_paid = models.DecimalField(
        _("amount paid"), decimal_places=4, max_digits=20, default=0
    )
    refunded_on = models.DateTimeField(
        _("refunded on"), blank=True, null=True, default=None, db_index=True
    )
    amount_refunded = models.DecimalField(
        _("amount refunded"), decimal_places=4, max_digits=20, default=0
    )
    external_id = models.CharField(
        _("external id"), max_length=64, blank=True, db_index=True, default=""
    )
    description = models.CharField(
        _("description"), max_length=128, blank=True, default=""
    )
    fraud_status = models.CharField(
        _("fraud status"),
        max_length=20,
        choices=FraudStatus.CHOICES,
        default=FraudStatus.UNKNOWN,
        db_index=True,
    )
    fraud_message = models.TextField(_("fraud message"), blank=True)
    _processor = None

    class Meta:
        abstract = True
        ordering = ["-created_on"]
        verbose_name = _("Payment")
        verbose_name_plural = _("Payments")

    def __str__(self):
        return "Payment #{self.id}".format(self=self)

    @property
    def processor(self):
        if self._processor is None:
            self._processor = self.get_processor()
        return self._processor

    def get_items(self):
        """
        Some backends require the list of items to be added to Payment.

        This method relays the call to Order. It is here simply because
        you can change the Order's fieldname when customizing Payment model.
        """
        return self.order.get_items()

    def get_processor(self):
        """
        Returns the processor instance for the backend that
        was chosen for this Payment. By default it takes it from global
        backend registry and tries to import it when it's not there.
        You most probably don't want to mess with this.
        """
        if self.backend in registry:
            processor = registry[self.backend]
        else:
            # last resort if backend has been removed from INSTALLED_APPS
            module = import_module(self.backend)
            processor = getattr(module, "PaymentProcessor")
        return processor(self)

    def change_status(self, new_status):
        """
        Used for changing the status of the Payment.
        You should always change payment status via this method.
        Otherwise the signal will not be emitted.
        """
        if self.status != new_status:
            # do anything only when status is really changed
            old_status = self.status
            self.status = new_status
            self.save()
            signals.payment_status_changed.send(
                sender=self.__class__,
                instance=self,
                old_status=old_status,
                new_status=new_status,
            )

    def change_fraud_status(self, new_status, message=""):
        """
        Used for changing fraud status of the Payment.
        You should always change payment fraud status via this method.
        Otherwise the signal will not be emitted.
        """
        if self.fraud_status != new_status:
            # do anything only when status is really changed
            old_status = self.fraud_status
            self.fraud_status = new_status
            self.fraud_message = message
            self.save()
            signals.payment_fraud_changed.send(
                sender=self.__class__,
                instance=self,
                old_status=old_status,
                new_status=new_status,
                message=message,
            )

    def on_success(self, amount=None):
        """
        Called when payment receives successful balance income. It defaults to
        complete payment, but can optionally accept received amount as a parameter
        to handle partial payments.

        Returns boolean value whether payment was fully paid.
        """
        if getattr(settings, "USE_TZ", False):
            timezone = getattr(settings, "TIME_ZONE", "local")
            self.paid_on = pendulum.now(timezone)
        else:
            self.paid_on = pendulum.now("UTC")

        if amount:
            self.amount_paid = amount
        else:
            self.amount_paid = self.amount

        fully_paid = self.amount_paid >= self.amount

        if fully_paid:
            self.change_status("paid")
        else:
            self.change_status("partially_paid")

        return fully_paid

    def on_failure(self):
        """
        Called when payment has failed. By default changes status to 'failed'
        """
        self.change_status("failed")

    def get_redirect_params(self):
        """
        Interfaces processor's ``get_redirect_params``.

        Redirect params is a dictionary containing all the data required by
        backend to process the payment in appropriate format.
        The data is extracted from Paymentand Order.
        """
        return self.processor.get_redirect_params()

    def get_redirect_url(self):
        """
        Interfaces processor's ``get_redirect_url``.

        Returns URL where the user will be redirected to complete the payment.
        """
        return self.processor.get_redirect_url()

    def get_redirect_method(self):
        """
        Interfaces processor's ``get_redirect_method``.

        Returns the method to be used to complete the payment - 'POST' or 'GET'.
        """
        return self.processor.get_redirect_method()

    def get_form(self, *args, **kwargs):
        """
        Interfaces processor's ``get_form``.

        Returns a Form to be used on intermediate page if the method returned by
        ``get_redirect_method`` is 'POST'.
        """
        return self.processor.get_form(*args, **kwargs)

    def get_template_names(self, view=None):
        """
        Interfaces processor's ``get_template_names``.

        Used to get templates for intermediate page when ``get_redirect_method``
        returns 'POST'.
        """
        return self.processor.get_template_names(view=view)

    def handle_callback(self, request, *args, **kwargs):
        """
        Interfaces processor's ``handle_callback``.

        Called when 'PUSH' flow is used for a backend. In this scenario
        broker's server will send a request to our server with information
        about the state of Payment. Broker can send several such requests during
        Payment's lifetime. Backend should analyze this request and return
        appropriate response that can be understood by broker's service.

        :param request: Request sent by payment broker.

        :return: HttpResponse instance
        """
        return self.processor.handle_callback(request, *args, **kwargs)

    def fetch_status(self):
        """
        Interfaces processor's ``fetch_status``.

        Used during 'PULL' flow. Fetches status from broker's service
        and translates it to a value from ``PAYMENT_STATUS_CHOICES``.
        """
        return self.processor.fetch_status()

    def fetch_and_update_status(self):
        """
        Used during 'PULL' flow to automatically fetch and update
        Payment's status.
        """
        remote_status = self.fetch_status()
        status = remote_status.get("status", None)
        amount = remote_status.get("amount", None)
        if (status is not None and "paid" in status) or amount is not None:
            self.on_success(amount)
        elif status == "failed":
            self.on_failure()
        elif status is not None:
            self.change_status(status)


class Payment(AbstractPayment):
    class Meta(AbstractPayment.Meta):
        swappable = swapper.swappable_setting("getpaid", "Payment")



from django.utils.translation import pgettext_lazy

__version__ = "2.0.0-rc.1"
default_app_config = "getpaid.apps.GetpaidConfig"


class PaymentStatus:
    NEW = "new"
    IN_PROGRESS = "in_progress"
    ACCEPTED = "accepted_for_proc"
    PAID = "paid"
    PARTIAL = "partially_paid"
    CANCELLED = "cancelled"
    FAILED = "failed"
    REFUNDED = "refunded"

    CHOICES = (
        (NEW, pgettext_lazy("payment status", "new")),
        (IN_PROGRESS, pgettext_lazy("payment status", "in progress")),
        (ACCEPTED, pgettext_lazy("payment status", "accepted for processing")),
        (PARTIAL, pgettext_lazy("payment status", "partially paid")),
        (PAID, pgettext_lazy("payment status", "paid")),
        (CANCELLED, pgettext_lazy("payment status", "cancelled")),
        (FAILED, pgettext_lazy("payment status", "failed")),
        (REFUNDED, pgettext_lazy("payment status", "refunded")),
    )

    values = [s[0] for s in CHOICES]


class FraudStatus:
    UNKNOWN = "unknown"
    ACCEPTED = "accepted"
    REJECTED = "rejected"
    CHECK = "check"

    CHOICES = (
        (UNKNOWN, pgettext_lazy("fraud status", "unknown")),
        (ACCEPTED, pgettext_lazy("fraud status", "accepted")),
        (REJECTED, pgettext_lazy("fraud status", "rejected")),
        (CHECK, pgettext_lazy("fraud status", "check")),
    )

    values = [s[0] for s in CHOICES]



import requests
from django.views.generic import FormView

from .forms import DummyQuestionForm


class DummyAuthorizationView(FormView):
    """
    This view simulates the behavior of payment broker
    """

    form_class = DummyQuestionForm
    template_name = "getpaid_dummy_backend/fake_gateway_authorization_form.html"
    success = None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        params = self.request.POST or self.request.GET  # both cases for testability
        context["payment"] = params.get("payment")
        context["value"] = params.get("value")
        context["currency"] = params.get("currency")
        context["description"] = params.get("description")
        context["callback"] = params.get("callback")
        context["success_url"] = params.get("success_url")
        context["failure_url"] = params.get("failure_url")
        return context

    def get_success_url(self):
        if self.success:
            return self.form.cleaned_data["success_url"]
        return self.form.cleaned_data["failure_url"]

    def form_valid(self, form):
        self.form = form
        url = self.request.build_absolute_uri(form.cleaned_data["callback"])
        # TODO: call post from delayed subprocess
        if form.cleaned_data["authorize_payment"] == "1":
            self.success = True
            requests.post(url, json={"status": "OK"})
        else:
            self.success = False
            requests.post(url, json={"status": "FAIL"})
        return super().form_valid(form)



import swapper
from django import http
from django.conf import settings
from django.core import exceptions
from django.shortcuts import get_object_or_404, resolve_url
from django.template.response import TemplateResponse
from django.utils.http import urlencode
from django.views import View
from django.views.generic import CreateView, RedirectView

from .forms import PaymentMethodForm


class CreatePaymentView(CreateView):
    model = swapper.load_model("getpaid", "Payment")
    form_class = PaymentMethodForm

    # template_name = "getpaid/payment_post_form.html"

    def get_form(self, form_class=None):
        if form_class is None:
            form_class = self.get_form_class()
        currency = self.kwargs["currency"]
        return form_class(currency=currency, **self.get_form_kwargs())

    def get(self, request, *args, **kwargs):
        """
        This view operates only on POST requests from order view where
        you select payment method
        """
        return http.HttpResponseNotAllowed(["POST"])

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        return kwargs

    def form_valid(self, form):
        payment = form.save()

        url = payment.get_redirect_url()
        method = payment.get_redirect_method()
        params = payment.get_redirect_params()
        payment.change_status("in_progress")
        if method.upper() == "GET":
            if params:
                url = "{url}?{params}".format(url=url, params=urlencode(params))
            return http.HttpResponseRedirect(url)
        elif method.upper() == "POST":
            context = self.get_context_data(
                form=payment.get_form(params), gateway_url=url
            )

            return TemplateResponse(
                request=self.request,
                # template=self.get_template_names(),
                template=payment.get_template_names(view=self),
                context=context,
            )
        else:
            raise exceptions.ImproperlyConfigured("Only GET and POST supported.")

    def form_invalid(self, form):
        raise exceptions.PermissionDenied


class FallbackView(RedirectView):
    """
    FallbackView (in form of either SuccessView or FailureView) handles the
    return from payment broker.
    """

    success = None
    permanent = False

    def get_redirect_url(self, *args, **kwargs):
        Payment = swapper.load_model("getpaid", "Payment")

        getpaid_settings = getattr(settings, "GETPAID", {})
        payment = get_object_or_404(Payment, pk=self.kwargs["pk"])
        if self.success:
            payment.on_success()
            url = getattr(getpaid_settings, "SUCCESS_URL", None)
        else:
            payment.on_failure()
            url = getattr(getpaid_settings, "FAILURE_URL", None)

        if url is not None:
            return resolve_url(
                url, pk=payment.order.pk
            )  # we may want to return to the Order summary or smth
        return resolve_url(
            payment.order.get_redirect_url(payment, success=self.success)
        )


class SuccessView(FallbackView):
    success = True


class FailureView(FallbackView):
    success = False


class CallbackView(View):
    def post(self, request, pk, *args, **kwargs):
        Payment = swapper.load_model("getpaid", "Payment")
        payment = get_object_or_404(Payment, pk=pk)
        return payment.handle_callback(request, *args, **kwargs)
